/*****************************************
 Copyright (c) 2002-2004
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/

/* This file is part of the boot loader 
   This driver for mrua_EM8620L_2.4.x.x */

/*
 * dm9102.c   V2.00
 *
 * driver for DM9102A/D PCI ethernet adapter
 *
 * by Spenser Tsai Nov. 2004
 * V1.01 - Spenser 21/12/2004
 *         . Fix filter setting.
 * V2.00 - Bill 02/06/2005
 *         . modify the parameters of dmfe_interrupt()
 */

#include "config.h"
#include "uart.h"
#include "util.h"
#include "net.h"
#include "net_ipv4.h"
#include "pcicommon.h"
#include "io.h"
#include "timer.h"

#define DM9102_VENDOR_ID	0x1282
#define DM9102_DEVICE_ID	0x9102

#define TX_DESC_CNT     0x4            /* Allocated Tx descriptors */
#define RX_DESC_CNT     0x10            /* Allocated Rx descriptors */
#define TX_FREE_DESC_CNT (TX_DESC_CNT - 2)	/* Max TX packet count */
#define TX_WAKE_DESC_CNT (TX_DESC_CNT - 3)	/* TX wakeup count */
#define DESC_ALL_CNT    (TX_DESC_CNT + RX_DESC_CNT)
#define TX_BUF_ALLOC    0x600
#define RX_ALLOC_SIZE   0x620
#define CR0_DEFAULT     0x00E00000      /* TX & RX burst mode */
#define CR6_DEFAULT     0x00080000      /* HD */
#define CR7_DEFAULT     0x180c1
#define CR15_DEFAULT    0x06            /* TxJabber RxWatchdog */

#define DMFE_10MHF      0
#define DMFE_100MHF     1
#define DMFE_10MFD      4
#define DMFE_100MFD     5
#define DMFE_AUTO       8
#define DMFE_1M_HPNA    0x10


/* CR9 definition: SROM/MII */
#define CR9_SROM_READ   0x4800
#define CR9_SRCS        0x1
#define CR9_SRCLK       0x2
#define CR9_CRDOUT      0x8
#define SROM_DATA_0     0x0
#define SROM_DATA_1     0x4
#define PHY_DATA_1      0x20000
#define PHY_DATA_0      0x00000
#define MDCLKH          0x10000

#define MAX_PACKET_SIZE 1514

#define TX_MAX_SEND_CNT 0x1
#define TDES0_ERR_MASK  0x4302          /* TXJT, LC, EC, FUE */


/* Structure/enum declaration ------------------------------- */
struct tx_desc {
        unsigned long tdes0, tdes1, tdes2, tdes3; /* Data for the card */
        unsigned long tx_buf_ptr;               /* Data for us */
        struct tx_desc *next_tx_desc;
} __attribute__(( aligned(32) ));

struct rx_desc {
	unsigned long rdes0, rdes1, rdes2, rdes3; /* Data for the card */	
	unsigned long data_buff;		// Data buffer
//	struct sk_buff *rx_skb_ptr;	/* Data for us */
	struct rx_desc *next_rx_desc;
} __attribute__(( aligned(32) ));

typedef struct dmfe_board_info {
	unsigned long chip_id;			/* Chip vendor/Device ID */
	unsigned long chip_revision;		/* Chip revision */
	struct DEVICE *next_dev;	/* next device */
	struct my_pci_dev pcidev;		/* PCI device */
//	spinlock_t lock;

	long ioaddr;			/* I/O base address */
	unsigned long cr0_data;
	unsigned long cr5_data;
	unsigned long cr6_data;
	unsigned long cr7_data;
	unsigned long cr15_data;

	/* pointer for memory physical address */
	unsigned long buf_pool_dma_ptr;	/* Tx buffer pool memory */
	unsigned long buf_pool_dma_start;	/* Tx buffer pool align dword */
	unsigned long desc_pool_dma_ptr;	/* descriptor pool memory */
	unsigned long first_tx_desc_dma;
	unsigned long first_rx_desc_dma;

	/* descriptor pointer */
	unsigned long buf_pool_ptr;	/* Tx buffer pool memory */
	unsigned long buf_pool_start;	/* Tx buffer pool align dword */
	unsigned long desc_pool_ptr;	/* descriptor pool memory */
	struct tx_desc *first_tx_desc;
	struct tx_desc *tx_insert_ptr;
	struct tx_desc *tx_remove_ptr;
	struct rx_desc *first_rx_desc;
	struct rx_desc *rx_insert_ptr;
	struct rx_desc *rx_ready_ptr;	/* packet come pointer */
	unsigned long tx_packet_cnt;	/* transmitted packet count */
	unsigned long tx_queue_cnt;	/* wait to send packet count */
	unsigned long rx_avail_cnt;	/* available rx descriptor count */
	unsigned long interval_rx_cnt;	/* rx packet count a callback time */

	unsigned int HPNA_command;		/* For HPNA register 16 */
	unsigned int HPNA_timer;			/* For HPNA remote device check */
	unsigned int dbug_cnt;
	unsigned int NIC_capability;		/* NIC media capability */
	unsigned int PHY_reg4;			/* Saved Phyxcer register 4 value */

	unsigned char HPNA_present;		/* 0:none, 1:DM9801, 2:DM9802 */
	unsigned char chip_type;			/* Keep DM9102A chip type */
	unsigned char media_mode;			/* user specify media mode */
	unsigned char op_mode;			/* real work media mode */
	unsigned char phy_addr;
	unsigned char link_failed;			/* Ever link failed */
	unsigned char wait_reset;			/* Hardware failed, need to reset */
	unsigned char dm910x_chk_mode;		/* Operating mode check */
	unsigned char first_in_callback;		/* Flag to record state */

	/* System defined statistic counter */

	/* Driver defined statistic counter */
	unsigned long tx_fifo_underrun;
	unsigned long tx_loss_carrier;
	unsigned long tx_no_carrier;
	unsigned long tx_late_collision;
	unsigned long tx_excessive_collision;
	unsigned long tx_jabber_timeout;
	unsigned long reset_count;
	unsigned long reset_cr8;
	unsigned long reset_fatal;
	unsigned long reset_TXtimeout;

	/* NIC SROM data */
	unsigned char srom[128];
} board_info_t;

enum dmfe_offsets {
	DCR0 = 0x00, DCR1 = 0x08, DCR2 = 0x10, DCR3 = 0x18, DCR4 = 0x20,
	DCR5 = 0x28, DCR6 = 0x30, DCR7 = 0x38, DCR8 = 0x40, DCR9 = 0x48,
	DCR10 = 0x50, DCR11 = 0x58, DCR12 = 0x60, DCR13 = 0x68, DCR14 = 0x70,
	DCR15 = 0x78
};

enum dmfe_CR6_bits {
	CR6_RXSC = 0x2, CR6_PBF = 0x8, CR6_PM = 0x40, CR6_PAM = 0x80,
	CR6_FDM = 0x200, CR6_TXSC = 0x2000, CR6_STI = 0x100000,
	CR6_SFT = 0x200000, CR6_RXA = 0x40000000, CR6_NO_PURGE = 0x20000000
};


// global configuration & variables
// 

static board_info_t g_board_info;
static unsigned char dmfe_media_mode = DMFE_AUTO;

//
// Basic I/O
//

#define dm9102_readb(reg)          pci_readb(g_board_info.ioaddr + (reg))
#define dm9102_readw(reg)          pci_readw(g_board_info.ioaddr + (reg))
#define dm9102_readl(reg)          pci_readl(g_board_info.ioaddr + (reg))

#define dm9102_writeb(val, reg)    pci_writeb(val, g_board_info.ioaddr + (reg))
#define dm9102_writew(val, reg)    pci_writew(val, g_board_info.ioaddr + (reg))
#define dm9102_writel(val, reg)    pci_writel(val, g_board_info.ioaddr + (reg))

#define SROM_CLK_WRITE(data) dm9102_writel(data|CR9_SROM_READ|CR9_SRCS,DCR9);delay(5);dm9102_writel(data|CR9_SROM_READ|CR9_SRCS|CR9_SRCLK,DCR9);delay(5);dm9102_writel(data|CR9_SROM_READ|CR9_SRCS,DCR9);delay(5);

//
// Function prototypes
//

int dm9102_probe(struct net_device *dev);
static int dmfe_init_dm910x(struct net_device *dev);
static int dmfe_open(struct net_device *dev);
static int dm9102_close(struct net_device *dev);
static int dmfe_start_xmit(struct sk_buff *skb, struct net_device *dev, int async);
static void dm9102_print_status(struct net_device *dev);
static void dmfe_interrupt(int irq,void *dev_id);
static unsigned int read_srom_word(int offset);
static void dmfe_descriptor_init(board_info_t *db, unsigned long ioaddr);
static void update_cr6(unsigned long cr6_data);
static void send_filter_frame(struct net_device *dev);
static void allocate_rx_buffer(board_info_t *db);
static void dmfe_rx_packet(struct net_device *dev, board_info_t * db);
static void dmfe_free_tx_pkt(struct net_device *dev, board_info_t * db);
static void dmfe_reuse_skb(board_info_t *db);
static int dm9102_receive_packet(struct net_device *dev);

//
// Init & Main
//

int dm9102_probe(struct net_device *dev)
{
    board_info_t *db = &g_board_info;

    if (pci_lookup(DM9102_VENDOR_ID, DM9102_DEVICE_ID, &db->pcidev)) {
        uart_printf("Found DM9102D(A) at PCI IDSEL %d\n", db->pcidev.idsel);

        // check I/O address 
        pci_configure_device(&db->pcidev);
        uart_printf("  Assign I/O address space from 0x%08x, size 0x%x\n", db->pcidev.resource[0].start, db->pcidev.resource[0].size);
        uart_printf("  Assign Memory address space from 0x%08x, size 0x%x\n", db->pcidev.resource[1].start, db->pcidev.resource[1].size);
        uart_printf("  Use Memory Mapped I/O\n");
        db->ioaddr = db->pcidev.resource[1].start;

        // setup Rx and Tx DES/buffer
	db->desc_pool_ptr = (unsigned long) malloc(sizeof(struct tx_desc) * DESC_ALL_CNT + 0x20);
	db->desc_pool_dma_ptr = virt_to_bus(db->desc_pool_ptr);	//Physical address
	db->buf_pool_ptr = (unsigned long) malloc(TX_BUF_ALLOC * TX_DESC_CNT + 4);	//Allocate TX buffer
	db->buf_pool_dma_ptr = virt_to_bus(db->buf_pool_ptr);//Physical address of RX buffer
		
	db->first_tx_desc = (struct tx_desc *) db->desc_pool_ptr;
	db->first_tx_desc_dma = db->desc_pool_dma_ptr;//Descriptor start address
	db->buf_pool_start = db->buf_pool_ptr;
	db->buf_pool_dma_start = db->buf_pool_dma_ptr;

        em86xx_flush_cache_data();
    
        // register functions
        dev->state = NETDEV_DOWN;
        dev->priv = db;
        dev->init = dmfe_init_dm910x;
        dev->open = dmfe_open;
        dev->close = dm9102_close;
        dev->send_packet = dmfe_start_xmit;
        dev->receive_packet = dm9102_receive_packet;
        dev->print_status = dm9102_print_status;

        return 1;
    }

    return 0;
}

int dm9102_close(struct net_device *dev)
{
    board_info_t *db = (board_info_t *) dev->priv;

    /* Disable TX/RX */
    dm9102_writel((db->cr6_data & ~0x2002), DCR6);
    em86xx_free_irq(db->pcidev.irqline);

    dev->state = NETDEV_DOWN;
    
    return 0;
}

int dmfe_open(struct net_device *dev)
{
    board_info_t *db = (board_info_t *) dev->priv;

    dmfe_init_dm910x(dev);

    em86xx_request_irq(db->pcidev.irqline, dmfe_interrupt, dev);

    dev->state = NETDEV_UP;

    return 0;
}


/*	Initilize DM910X board
 *	Reset DM910X board
 *	Initilize TX/Rx descriptor chain structure
 *	Send the set-up frame
 *	Enable Tx/Rx machine
 */

int dmfe_init_dm910x(struct net_device *dev)
{
	struct dmfe_board_info *db = (board_info_t *)dev->priv;
	unsigned long ioaddr = db->ioaddr;
	int i;
	
	/* system variable init */
	db->cr6_data = CR6_DEFAULT;
	db->tx_packet_cnt = 0;
	db->tx_queue_cnt = 0;
	db->rx_avail_cnt = 0;
	//db->link_failed = 1;
	db->wait_reset = 0;

	//db->first_in_callback = 0;
	//db->NIC_capability = 0xf;	/* All capability*/
	//db->PHY_reg4 = 0x1e0;

	/* Reset DM910x MAC controller */
	dm9102_writel(0x1, DCR0);	/* RESET MAC */
	delay(100);
	dm9102_writel(db->cr0_data, DCR0);
	delay(5);

	/* Phy addr : DM910(A)2/DM9132/9801, phy address = 1 */
	//db->phy_addr = 1;


	/* Set Node address */
	((unsigned short *) dev->dev_addr)[0] = read_srom_word(10);
	((unsigned short *) dev->dev_addr)[1] = read_srom_word(11);
	((unsigned short *) dev->dev_addr)[2] = read_srom_word(12);

   	uart_puts("  MAC address is ");
    	for (i = 0; i < MACADDR_LEN; ++i) {
        	uart_printf("%02X", dev->dev_addr[i]);
        	if (i < MACADDR_LEN - 1)
            		uart_puts(":");
    	}
    	uart_puts("\n");

	db->media_mode = dmfe_media_mode;
	db->op_mode = db->media_mode; 	/* Force Mode */

	/* RESET Phyxcer Chip by GPR port bit 7 */
	dm9102_writel(0x180, DCR12);	/* Let bit 7 output port */
	dm9102_writel(0x0, DCR12);	/* Clear RESET signal */


	/* Media Mode Process */
	//if ( !(db->media_mode & DMFE_AUTO) )
	//	db->op_mode = db->media_mode; 	/* Force Mode */

	/* Initiliaze Transmit/Receive decriptor and CR3/4 */
	dmfe_descriptor_init(db, ioaddr);

	/* Init CR6 to program DM910x operation */
	update_cr6(db->cr6_data);

	/* Send setup frame */
	send_filter_frame(dev);	/* DM9102/DM9102A */

	/* Init CR7, interrupt active bit */
	db->cr7_data = CR7_DEFAULT;
	dm9102_writel(db->cr7_data, DCR7);

	/* Init CR15, Tx jabber and Rx watchdog timer */
	dm9102_writel(db->cr15_data, DCR15);

	/* Enable DM910X Tx/Rx function */
	db->cr6_data |= CR6_RXSC | CR6_TXSC | 0x40000;
	update_cr6(db->cr6_data);
	
	return 0;
}

/*
 *	Read one word data from the serial ROM
 */
unsigned int read_srom_word(int offset)
{
	int i;
	unsigned int srom_data = 0;

	dm9102_writel(CR9_SROM_READ, DCR9);
	dm9102_writel(CR9_SROM_READ | CR9_SRCS, DCR9);

	/* Send the Read Command 110b */
	SROM_CLK_WRITE(SROM_DATA_1);
	SROM_CLK_WRITE(SROM_DATA_1);
	SROM_CLK_WRITE(SROM_DATA_0);

	/* Send the offset */
	for (i = 5; i >= 0; i--) {
		srom_data = (offset & (1 << i)) ? SROM_DATA_1 : SROM_DATA_0;
		SROM_CLK_WRITE(srom_data);
	}

	dm9102_writel(CR9_SROM_READ | CR9_SRCS, DCR9);

	for (i = 16; i > 0; i--) {
		dm9102_writel(CR9_SROM_READ | CR9_SRCS | CR9_SRCLK, DCR9);
		delay(5);
		srom_data = (srom_data << 1) | ((dm9102_readl(DCR9) & CR9_CRDOUT) ? 1 : 0);
		dm9102_writel(CR9_SROM_READ | CR9_SRCS, DCR9);
		delay(5);
	}

	dm9102_writel(CR9_SROM_READ, DCR9);
	return srom_data;
}

/*
 *	Initialize transmit/Receive descriptor
 *	Using Chain structure, and allocate Tx/Rx buffer
 */

void dmfe_descriptor_init(board_info_t *db, unsigned long ioaddr)
{
	struct tx_desc *tmp_tx;
	struct rx_desc *tmp_rx;
	unsigned long tmp_buf;
	unsigned long tmp_tx_dma, tmp_rx_dma;
	unsigned long tmp_buf_dma;
	int i;

	/* tx descriptor start pointer */
	db->tx_insert_ptr = db->first_tx_desc;
	db->tx_remove_ptr = db->first_tx_desc;
	dm9102_writel(db->first_tx_desc_dma, DCR4);     /* TX DESC address */

	/* rx descriptor start pointer */
	db->first_rx_desc = (void *)db->first_tx_desc + sizeof(struct tx_desc) * TX_DESC_CNT;
	db->first_rx_desc_dma =  db->first_tx_desc_dma + sizeof(struct tx_desc) * TX_DESC_CNT;
	db->rx_insert_ptr = db->first_rx_desc;
	db->rx_ready_ptr = db->first_rx_desc;
	dm9102_writel(db->first_rx_desc_dma, DCR3);	/* RX DESC address */

	/* Init Transmit chain */
	tmp_buf = db->buf_pool_start;
	tmp_buf_dma = db->buf_pool_dma_start;
	tmp_tx_dma = db->first_tx_desc_dma;
	for (tmp_tx = db->first_tx_desc, i = 0; i < TX_DESC_CNT; i++, tmp_tx++) 	{
		tmp_tx->tx_buf_ptr = tmp_buf;
		tmp_tx->tdes0 = 0;
		tmp_tx->tdes1 = 0x81000000;// IC, chain, Disable Checksum offload
		tmp_tx->tdes2 = tmp_buf_dma;
		tmp_tx_dma += sizeof(struct tx_desc);
		tmp_tx->tdes3 = tmp_tx_dma;
		tmp_tx->next_tx_desc = tmp_tx + 1;
		tmp_buf = tmp_buf + TX_BUF_ALLOC;
		tmp_buf_dma = tmp_buf_dma + TX_BUF_ALLOC;
	}
	(--tmp_tx)->tdes3 = db->first_tx_desc_dma;//Chain to 1st DESC.
	tmp_tx->next_tx_desc = db->first_tx_desc;//Empty DESC.

	 /* Init Receive descriptor chain */
	tmp_rx_dma=db->first_rx_desc_dma;
	for (tmp_rx = db->first_rx_desc, i = 0; i < RX_DESC_CNT; i++, tmp_rx++) 	{
		tmp_rx->rdes0 = 0;
		tmp_rx->rdes1 = 0x01000600;
		tmp_rx_dma += sizeof(struct rx_desc);
		tmp_rx->rdes3 = tmp_rx_dma;
		tmp_rx->next_rx_desc = tmp_rx + 1;
	}
	(--tmp_rx)->rdes3 = db->first_rx_desc_dma;
	tmp_rx->next_rx_desc = db->first_rx_desc;

	/* pre-allocate Rx buffer */
	allocate_rx_buffer(db);
}

/*
 *	Allocate rx buffer,
 *	As possible as allocate maxiumn Rx buffer
 */

void allocate_rx_buffer(board_info_t *db)
{
	struct rx_desc *rxptr;

	rxptr = db->rx_insert_ptr;

	while(db->rx_avail_cnt < RX_DESC_CNT) {
		rxptr->data_buff = (unsigned long) malloc(RX_ALLOC_SIZE); 
		rxptr->rdes2 = virt_to_bus((unsigned long)rxptr->data_buff);
		rxptr->rdes0 = 0x80000000;
		rxptr = rxptr->next_rx_desc;
		db->rx_avail_cnt++;
	}

	db->rx_insert_ptr = rxptr;
}


/*
 *	Update CR6 value
 *	Firstly stop DM910X , then written value and start
 */

void update_cr6(unsigned long cr6_data)
{
	unsigned long cr6_tmp;

	cr6_tmp = cr6_data & ~0x2002;           /* stop Tx/Rx */
	dm9102_writel(cr6_tmp, DCR6);
	delay(5);
	dm9102_writel(cr6_data, DCR6);
	delay(5);
}


/*
 *	Send a setup frame for DM9102/DM9102A
 *	This setup frame initilize DM910X addres filter mode
 */

void send_filter_frame(struct net_device *dev)
{
	board_info_t *db = dev->priv;
	struct tx_desc *txptr;
	unsigned char * addrptr;
	unsigned long * suptr;
	
	txptr = db->tx_insert_ptr;
	suptr = (unsigned long *) txptr->tx_buf_ptr;
	
	/* Node address */
	addrptr = (unsigned char *) dev->dev_addr;
	*suptr++ = (unsigned long)addrptr[0] | ((unsigned long)addrptr[1] << 8);
	*suptr++ = (unsigned long)addrptr[2] | ((unsigned long)addrptr[3] << 8);
	*suptr++ = (unsigned long)addrptr[4] | ((unsigned long)addrptr[5] << 8);
	/* broadcast address */
	*suptr++ = 0xffff;
	*suptr++ = 0xffff;
	*suptr++ = 0xffff;

	/* prepare the setup frame */
	db->tx_insert_ptr = txptr->next_tx_desc;
	txptr->tdes1 = 0x890000c0;

	/* Resource Check and Send the setup packet */
	if (!db->tx_packet_cnt) {
		/* Resource Empty */
		db->tx_packet_cnt++;
		txptr->tdes0 = 0x80000000;
		update_cr6(db->cr6_data | 0x2000);
		dm9102_writel(0x1, DCR1);	/* Issue Tx polling */
		update_cr6(db->cr6_data);
	} else
		db->tx_queue_cnt++;	/* Put in TX queue */
}


/*
 *	Hardware start transmission.
 *	Send a packet to media from the upper layer.
 */

int dmfe_start_xmit(struct sk_buff *skb, struct net_device *dev, int async)
{
	struct dmfe_board_info *db = dev->priv;
	struct tx_desc *txptr;

	/* Too large packet check */
	if (skb->len > MAX_PACKET_SIZE) {
		uart_printf("<DM9102>: big packet = %d\n", (unsigned int)skb->len);
		return 1;
	}

	/* Disable NIC interrupt */
	dm9102_writel(0, DCR7);

	/* transmit this packet */
	txptr = db->tx_insert_ptr;
	memcpy((void *)txptr->tx_buf_ptr, skb->data, skb->len);
    	em86xx_clean_cache_data_region(txptr->tx_buf_ptr, txptr->tx_buf_ptr + skb->len);

	txptr->tdes1 = 0xe1000000 | skb->len;

	/* Point to next transmit free descriptor */
	db->tx_insert_ptr = txptr->next_tx_desc;

	/* Transmit Packet Process */
	if ( (!db->tx_queue_cnt) && (db->tx_packet_cnt < TX_MAX_SEND_CNT) ) {
		txptr->tdes0 = 0x80000000;	/* Set owner bit */
		db->tx_packet_cnt++;			/* Ready to send */
		dm9102_writel(0x1, DCR1);	/* Issue Tx polling */
	} else {
		db->tx_queue_cnt++;			/* queue TX packet */
		dm9102_writel(0x1, DCR1);	/* Issue Tx polling */
	}

//	if (!async)
//		uart_printf("<DM9102>: In synchronous mode, it must disable interrupt mode.\n");

	/* free this SKB */
	skb_free(skb);

	/* Restore CR7 to enable interrupt */
	dm9102_writel(db->cr7_data, DCR7);

	return 0;
}

void dm9102_print_status(struct net_device *dev)
{
//	uart_printf("dm9102_print_status()....\n");
}

/*
 *	DM9102 insterrupt handler
 *	receive the packet to upper layer, free the transmitted packet
 */

void dmfe_interrupt(int irq ,void *dev_id)
{
	struct net_device *dev = dev_id;
	board_info_t *db = (board_info_t *) dev->priv;

	if (!dev) {
		return;
	}

	/* Got DM910X status */
	db->cr5_data = dm9102_readl(DCR5);
	dm9102_writel(db->cr5_data, DCR5);
	if ( !(db->cr5_data & 0xc1) ) {
//		uart_puts("SST: CR5 is not 0xC1...\n");
		return;
	}

	/* Disable all interrupt in CR7 to solve the interrupt edge problem */
	dm9102_writel(0, DCR7);

	/* Check system status */
	if (db->cr5_data & 0x2000) {
		/* system bus error happen */
//		uart_printf("SST: System bus error happen. CR5= 0x%08X\n", db->cr5_data);
		db->reset_fatal++;
		db->wait_reset = 1;	/* Need to RESET */
		return;
	}

	 /* Received the coming packet */
	if ( (db->cr5_data & 0x40) && db->rx_avail_cnt )
		dmfe_rx_packet(dev, db);


	/* Free the transmitted descriptor */
	if ( db->cr5_data & 0x01)
		dmfe_free_tx_pkt(dev, db);

	/* Mode Check */
	if (db->dm910x_chk_mode & 0x2) {
		db->dm910x_chk_mode = 0x4;
		db->cr6_data |= 0x100;
		update_cr6(db->cr6_data);
	}

	/* Restore CR7 to enable interrupt mask */
	dm9102_writel(db->cr7_data, DCR7);

}


/*
 *	Receive the come packet and pass to upper layer
 */

void dmfe_rx_packet(struct net_device *dev, board_info_t * db)
{
	struct rx_desc *rxptr;
	struct sk_buff *skb;
	int rxlen;
	unsigned long rdes0;

//	uart_printf(SST: "dmfe_rx_packet().........\n");

	rxptr = db->rx_ready_ptr;
	
	em86xx_flush_cache_all();

	while(db->rx_avail_cnt) {
		rdes0 = rxptr->rdes0;
		if (rdes0 & 0x80000000)	/* packet owner check */
			break;

		db->rx_avail_cnt--;
		db->interval_rx_cnt++;

		if ( (rdes0 & 0x300) != 0x300) {
			/* A packet without First/Last flag */
			/* reuse this SKB */
			dmfe_reuse_skb(db);
		} else {
			/* A packet with First/Last flag */
			rxlen = ( (rdes0 >> 16) & 0x3fff) - 4;

			if ( !(rdes0 & 0x8000) ||
				((db->cr6_data & CR6_PM) && (rxlen>6)) ) {
						if ((skb = skb_alloc(rxlen + 2)) != NULL) {

							skb->len = rxlen;
							memcpy(skb->data, rxptr->data_buff, rxlen);
							skb_put(skb);
						}
						dmfe_reuse_skb(db);
			} else {
				/* Reuse SKB buffer when the packet is error */
				dmfe_reuse_skb(db);
			}
		}

		rxptr = rxptr->next_rx_desc;
	}

	db->rx_ready_ptr = rxptr;
}


/*
 *	Reuse the SK buffer
 */

void dmfe_reuse_skb(board_info_t *db)
{
	struct rx_desc *rxptr = db->rx_insert_ptr;

	if (!(rxptr->rdes0 & 0x80000000)) {
		rxptr->rdes0 = 0x80000000;
		db->rx_avail_cnt++;
		db->rx_insert_ptr = rxptr->next_rx_desc;
	} 
}

/*
 *	Free TX resource after TX complete
 */

void dmfe_free_tx_pkt(struct net_device *dev, board_info_t * db)
{
	struct tx_desc *txptr;
	unsigned long tdes0;

//	uart_printf(SST: "dmfe_free_tx_pkt().........\n");

	txptr = db->tx_remove_ptr;
	while(db->tx_packet_cnt) {
		tdes0 = txptr->tdes0;
		/* printk(DRV_NAME ": tdes0=%x\n", tdes0); */
		if (tdes0 & 0x80000000)
			break;

		/* A packet sent completed */
		db->tx_packet_cnt--;

		/* Transmit statistic counter */
		if ( tdes0 != 0x7fffffff ) {
			/* printk(DRV_NAME ": tdes0=%x\n", tdes0); */
			if (tdes0 & TDES0_ERR_MASK) {

				if (tdes0 & 0x0002) {	/* UnderRun */
					db->tx_fifo_underrun++;
					if ( !(db->cr6_data & CR6_SFT) ) {	//Change to Store and forward
						db->cr6_data = db->cr6_data | CR6_SFT;
						update_cr6(db->cr6_data);
					}
				}
				if (tdes0 & 0x0100)
					db->tx_excessive_collision++;
				if (tdes0 & 0x0200)
					db->tx_late_collision++;
				if (tdes0 & 0x0400)
					db->tx_no_carrier++;
				if (tdes0 & 0x0800)
					db->tx_loss_carrier++;
				if (tdes0 & 0x4000)
					db->tx_jabber_timeout++;
			}
		}

    		txptr = txptr->next_tx_desc;
	}/* End of while */

	/* Update TX remove pointer to next */
	db->tx_remove_ptr = txptr;

	/* Send the Tx packet in queue */
	if ( (db->tx_packet_cnt < TX_MAX_SEND_CNT) && db->tx_queue_cnt ) {
		txptr->tdes0 = 0x80000000;	/* Set owner bit */
		db->tx_packet_cnt++;			/* Ready to send */
		db->tx_queue_cnt--;
		dm9102_writel(0x1, DCR1);		/* Issue Tx polling */
	}

}

int dm9102_receive_packet(struct net_device *dev)
{
    board_info_t *db = (board_info_t *) dev->priv;
    
//    uart_printf("SST: dm9102_receive_packet()..........\n");
    
    em86xx_mask_irq(db->pcidev.irqline);

    dmfe_interrupt(0,dev);

    em86xx_unmask_irq(db->pcidev.irqline);
    return 0;
}
